# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

import argparse
import os
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, TextIO, Union

import pandas as pd
from pydantic import Field
from yaml import safe_load

from gems.utils import ModifiedBaseModel


def parse_yaml_components(input_study: TextIO) -> "InputSystem":
    tree = safe_load(input_study)
    return InputSystem.model_validate(tree["system"])


def parse_scenario_builder(file: Path) -> pd.DataFrame:
    sb = pd.read_csv(file, names=("name", "year", "scenario"))
    sb.rename(columns={0: "name", 1: "year", 2: "scenario"})
    return sb


class InputAreaConnections(ModifiedBaseModel):
    component: str
    port: str
    area: str


class InputPortConnections(ModifiedBaseModel):
    component1: str
    port1: str
    component2: str
    port2: str


class InputComponentParameter(ModifiedBaseModel):
    id: str
    time_dependent: bool = False
    scenario_dependent: bool = False
    value: Union[float, str]
    scenario_group: Optional[str] = None


class InputComponent(ModifiedBaseModel):
    id: str
    model: str
    scenario_group: Optional[str] = None
    parameters: Optional[List[InputComponentParameter]] = None


class InputSystem(ModifiedBaseModel):
    id: Optional[str] = None
    model_libraries: Optional[str] = None  # Parsed but unused for now
    components: List[InputComponent] = Field(default_factory=list)
    connections: Optional[List[InputPortConnections]] = None
    area_connections: Optional[List[InputAreaConnections]] = None
    nodes: Optional[List[InputComponent]] = []


@dataclass(frozen=True)
class ParsedArguments:
    models_path: List[Path]
    components_path: Path
    timeseries_path: Path
    duration: int
    nb_scenarios: int


def parse_cli() -> ParsedArguments:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--study", type=Path, help="path to the root directory of the study"
    )
    parser.add_argument(
        "--models", nargs="+", type=Path, help="list of path to model file, *.yml"
    )
    parser.add_argument(
        "--component", type=Path, help="path to the component file, *.yml"
    )
    parser.add_argument(
        "--timeseries", type=Path, help="path to the timeseries directory"
    )
    parser.add_argument(
        "--duration", type=int, help="duration of the simulation", default=1
    )
    parser.add_argument(
        "--scenario", type=int, help="number of scenario of the simulation", default=1
    )

    args = parser.parse_args()

    if args.study:
        if args.models or args.component or args.timeseries:
            parser.error(
                "--study flag can't be use with --models, --component and --timeseries"
            )

        components_path = args.study / "input" / "components" / "components.yml"
        timeseries_dir = args.study / "input" / "components" / "series"
        model_paths = [
            args.study / "input" / "models" / file
            for file in os.listdir(args.study / "input" / "models")
        ]

    else:
        if not args.models or not args.component:
            parser.error("--models and --component must be entered")

        components_path = args.component
        timeseries_dir = args.timeseries
        model_paths = args.models

    return ParsedArguments(
        model_paths, components_path, timeseries_dir, args.duration, args.scenario
    )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
Module for technical utilities.
"""
import json
import pathlib
from typing import Any, Callable, Dict, Optional, TypeVar

from pydantic import BaseModel

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
Supplier = Callable[[], T]


def require_not_none(obj: Any, msg: Optional[str] = None) -> None:
    """
    Raises a ValueError if obj is None.
    """
    if obj is None:
        error_msg = msg if msg else "Object must not be None"
        raise ValueError(error_msg)


def get_or_add(dictionary: Dict[K, V], key: K, default_factory: Supplier[V]) -> V:
    """
    Gets value from dictionary, or inserts it if it does not exist.

    Factory is only called if value is absent.
    """
    value = dictionary.get(key)
    if not value:
        value = default_factory()
        dictionary[key] = value
    return value


def serialize(filename: str, message: str, path: pathlib.Path) -> None:
    """
    Write message to path/filename

    Will throw an exception if it fails to create dir or ro open file
    """
    path.mkdir(parents=True, exist_ok=True)
    file = (path / filename).open(mode="w")

    with file:
        file.write(message)


def serialize_json(
    filename: str, message: Dict[str, Any], path: pathlib.Path, indentation: int = 4
) -> None:
    serialize(filename, json.dumps(message, indent=indentation), path)


def read_json(filename: str, path: pathlib.Path) -> Dict[str, Any]:
    with (path / filename).open() as file:
        data = json.load(file)
    return data


"""
Only _to_kebab adn ModifiedBaseModel are used in the project.
"""


# Design note: actual parsing and validation is delegated to pydantic models
def _to_kebab(snake: str) -> str:
    return snake.replace("_", "-")


class ModifiedBaseModel(BaseModel):
    class Config:
        alias_generator = _to_kebab
        extra = "forbid"
        populate_by_name = True



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from gems.expression import (
    AdditionNode,
    DivisionNode,
    ExpressionVisitor,
    MultiplicationNode,
    NegationNode,
)
from gems.expression.expression import (
    AllTimeSumNode,
    ComparisonNode,
    ComponentParameterNode,
    ComponentVariableNode,
    CurrentScenarioIndex,
    ExpressionNode,
    LiteralNode,
    NoScenarioIndex,
    NoTimeIndex,
    OneScenarioIndex,
    ParameterNode,
    PortFieldAggregatorNode,
    PortFieldNode,
    ProblemParameterNode,
    ProblemVariableNode,
    ScenarioIndex,
    ScenarioOperatorNode,
    TimeEvalNode,
    TimeIndex,
    TimeShift,
    TimeShiftNode,
    TimeStep,
    TimeSumNode,
    VariableNode,
)
from gems.expression.visitor import visit
from gems.simulation.linear_expression import LinearExpression, Term, TermKey


class ParameterGetter(ABC):
    @abstractmethod
    def get_parameter_value(
        self,
        component_id: str,
        parameter_name: str,
        timestep: Optional[int],
        scenario: Optional[int],
    ) -> float:
        pass


@dataclass
class MutableTerm:
    coefficient: float
    component_id: str
    variable_name: str
    time_index: Optional[int]
    scenario_index: Optional[int]

    def to_key(self) -> TermKey:
        return TermKey(
            self.component_id,
            self.variable_name,
            self.time_index,
            self.scenario_index,
        )

    def to_term(self) -> Term:
        return Term(
            self.coefficient,
            self.component_id,
            self.variable_name,
            self.time_index,
            self.scenario_index,
        )


@dataclass
class LinearExpressionData:
    terms: List[MutableTerm]
    constant: float

    def build(self) -> LinearExpression:
        res_terms: Dict[TermKey, Any] = {}
        for t in self.terms:
            k = t.to_key()
            if k in res_terms:
                current_t = res_terms[k]
                current_t.coefficient += t.coefficient
            else:
                res_terms[k] = t
        for k, v in res_terms.items():
            res_terms[k] = v.to_term()
        return LinearExpression(res_terms, self.constant)


@dataclass(frozen=True)
class LinearExpressionBuilder(ExpressionVisitor[LinearExpressionData]):
    """
    Reduces a generic expression to a linear expression.

    The input expression must respect the constraints of the output of
    the operators expansion expression:
    it must only contain `ProblemVariableNode` for variables
    and `ProblemParameterNode` parameters. It cannot contain anymore
    time aggregators or scenario aggregators, nor port-related nodes.
    """

    # TODO: linear expressions should be re-usable for different timesteps and scenarios
    timestep: Optional[int]
    scenario: Optional[int]
    value_provider: Optional[ParameterGetter] = None

    def negation(self, node: NegationNode) -> LinearExpressionData:
        operand = visit(node.operand, self)
        operand.constant = -operand.constant
        for t in operand.terms:
            t.coefficient = -t.coefficient
        return operand

    def addition(self, node: AdditionNode) -> LinearExpressionData:
        operands = [visit(o, self) for o in node.operands]
        terms = []
        constant: float = 0
        for o in operands:
            constant += o.constant
            terms.extend(o.terms)
        return LinearExpressionData(terms=terms, constant=constant)

    def multiplication(self, node: MultiplicationNode) -> LinearExpressionData:
        lhs = visit(node.left, self)
        rhs = visit(node.right, self)
        if not lhs.terms:
            multiplier = lhs.constant
            actual_expr = rhs
        elif not rhs.terms:
            multiplier = rhs.constant
            actual_expr = lhs
        else:
            raise ValueError(
                "At least one operand of a multiplication must be a constant expression."
            )
        actual_expr.constant *= multiplier
        for t in actual_expr.terms:
            t.coefficient *= multiplier
        return actual_expr

    def division(self, node: DivisionNode) -> LinearExpressionData:
        lhs = visit(node.left, self)
        rhs = visit(node.right, self)
        if rhs.terms:
            raise ValueError(
                "The second operand of a division must be a constant expression."
            )
        divider = rhs.constant
        actual_expr = lhs
        actual_expr.constant /= divider
        for t in actual_expr.terms:
            t.coefficient /= divider
        return actual_expr

    def _get_timestep(self, time_index: TimeIndex) -> Optional[int]:
        if isinstance(time_index, TimeShift):
            if self.timestep is None:
                raise ValueError("Cannot shift a time-independent expression.")
            return self.timestep + time_index.timeshift
        if isinstance(time_index, TimeStep):
            return time_index.timestep
        if isinstance(time_index, NoTimeIndex):
            return None
        else:
            raise TypeError(f"Type {type(time_index)} is not a valid TimeIndex type.")

    def _get_scenario(self, scenario_index: ScenarioIndex) -> Optional[int]:
        if isinstance(scenario_index, OneScenarioIndex):
            return scenario_index.scenario
        elif isinstance(scenario_index, CurrentScenarioIndex):
            return self.scenario
        elif isinstance(scenario_index, NoScenarioIndex):
            return None
        else:
            raise TypeError(
                f"Type {type(scenario_index)} is not a valid ScenarioIndex type."
            )

    def literal(self, node: LiteralNode) -> LinearExpressionData:
        return LinearExpressionData([], node.value)

    def comparison(self, node: ComparisonNode) -> LinearExpressionData:
        raise ValueError("Linear expression cannot contain a comparison operator.")

    def variable(self, node: VariableNode) -> LinearExpressionData:
        raise ValueError(
            "Variables need to be associated with their component ID before linearization."
        )

    def parameter(self, node: ParameterNode) -> LinearExpressionData:
        raise ValueError("Parameters must be evaluated before linearization.")

    def comp_variable(self, node: ComponentVariableNode) -> LinearExpressionData:
        raise ValueError(
            "Variables need to be associated with their timestep/scenario before linearization."
        )

    def pb_variable(self, node: ProblemVariableNode) -> LinearExpressionData:
        return LinearExpressionData(
            [
                MutableTerm(
                    1,
                    node.component_id,
                    node.name,
                    time_index=self._get_timestep(node.time_index),
                    scenario_index=self._get_scenario(node.scenario_index),
                )
            ],
            0,
        )

    def comp_parameter(self, node: ComponentParameterNode) -> LinearExpressionData:
        raise ValueError(
            "Parameters need to be associated with their timestep/scenario before linearization."
        )

    def pb_parameter(self, node: ProblemParameterNode) -> LinearExpressionData:
        # TODO SL: not the best place to do this.
        # in the future, we should evaluate coefficients of variables as time vectors once for all timesteps
        time_index = self._get_timestep(node.time_index)
        scenario_index = self._get_scenario(node.scenario_index)
        return LinearExpressionData(
            [],
            self._value_provider().get_parameter_value(
                node.component_id, node.name, time_index, scenario_index
            ),
        )

    def time_eval(self, node: TimeEvalNode) -> LinearExpressionData:
        raise ValueError("Time operators need to be expanded before linearization.")

    def time_shift(self, node: TimeShiftNode) -> LinearExpressionData:
        raise ValueError("Time operators need to be expanded before linearization.")

    def time_sum(self, node: TimeSumNode) -> LinearExpressionData:
        raise ValueError("Time operators need to be expanded before linearization.")

    def all_time_sum(self, node: AllTimeSumNode) -> LinearExpressionData:
        raise ValueError("Time operators need to be expanded before linearization.")

    def _value_provider(self) -> ParameterGetter:
        if self.value_provider is None:
            raise ValueError(
                "A value provider must be specified to linearize a time operator node."
                " This is required in order to evaluate the value of potential parameters"
                " used to specified the time ids on which the time operator applies."
            )
        return self.value_provider

    def scenario_operator(self, node: ScenarioOperatorNode) -> LinearExpressionData:
        raise ValueError("Scenario operators need to be expanded before linearization.")

    def port_field(self, node: PortFieldNode) -> LinearExpressionData:
        raise ValueError("Port fields must be replaced before linearization.")

    def port_field_aggregator(
        self, node: PortFieldAggregatorNode
    ) -> LinearExpressionData:
        raise ValueError(
            "Port fields aggregators must be replaced before linearization."
        )


def linearize_expression(
    expression: ExpressionNode,
    timestep: Optional[int],
    scenario: Optional[int],
    value_provider: Optional[ParameterGetter] = None,
) -> LinearExpression:
    return visit(
        expression,
        LinearExpressionBuilder(
            value_provider=value_provider, timestep=timestep, scenario=scenario
        ),
    ).build()



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

import math
from dataclasses import dataclass, field
from typing import Dict, Generator, Iterable, List, Optional, Set

from anytree import LevelOrderIter, NodeMixin

from gems.expression import ExpressionNode, literal, var
from gems.expression.indexing_structure import IndexingStructure
from gems.model.common import ProblemContext
from gems.model.constraint import Constraint
from gems.model.model import model
from gems.model.variable import Variable, float_variable
from gems.simulation.time_block import TimeBlock
from gems.study.network import Component, Network, create_component


@dataclass(frozen=True)
class InterDecisionTimeScenarioConfig:
    blocks: List[TimeBlock]
    scenarios: int


@dataclass
class CouplingInfo:
    variables: Set[str] = field(default_factory=set)
    constraints: Dict[str, ExpressionNode] = field(default_factory=dict)


class DecisionTreeNode(NodeMixin):
    id: str
    config: InterDecisionTimeScenarioConfig
    network: Network
    prob: float

    coupling_network: Network
    coupling_info: CouplingInfo

    def __init__(
        self,
        id: str,
        config: InterDecisionTimeScenarioConfig,
        network: Network,
        parent: Optional["DecisionTreeNode"] = None,
        children: Optional[Iterable["DecisionTreeNode"]] = None,
        prob: float = 1.0,
        coupling_network: Network = Network("_Coupler"),
        coupling_info: CouplingInfo = CouplingInfo(),
    ) -> None:
        self.id = id
        self.config = config
        self.network = network
        self.coupling_network = coupling_network
        self.coupling_info = coupling_info
        self.parent = parent

        if prob < 0 or 1 < prob:
            raise ValueError("Probability must be a value in the range [0, 1]")

        self.prob = prob * (parent.prob if parent is not None else 1)
        if children:
            self.children = children

    def traverse(
        self, depth: Optional[int] = None
    ) -> Generator["DecisionTreeNode", None, None]:
        yield from LevelOrderIter(self, maxlevel=depth)

    def is_leaves_prob_sum_one(self) -> bool:
        if not self.children:
            return True

        # Since we multiply the child's prob by the parent's prob
        # in the constructor, the sum of the children prob should
        # equal 1 * parent.prob if the values were set correctly
        if not math.isclose(self.prob, sum(child.prob for child in self.children)):
            return False

        # Recursively check if child nodes have their children's
        # probability sum equal to one
        return all(child.is_leaves_prob_sum_one() for child in self.children)

    def define_coupling_constraint(
        self,
        component_par: Component,
        cumulative_par_var_id: str,
        component_chd: Component,
        cumulative_chd_var_id: str,
        incremental_chd_var_id: str,
    ) -> None:
        """
        Define a coupling constraint of the form:

        var(cumulative_par_var_id) == var(cumulative_chd_var_id) - var(incremental_chd_var_id)

        where 'cumulative_par_var_id' is a variable of component_par;\n
        and 'cumulative_chd_var_id' and 'incremental_chd_var_id' are variables of component_chd
        """
        if self.parent and (component_par not in self.parent.network.all_components):
            raise RuntimeError(
                f"Component {component_par.id} not present in parent's network!"
            )

        if component_chd not in self.network.all_components:
            raise RuntimeError(f"Component {component_chd.id} not present in network!")

        if not component_par.is_variable_in_model(cumulative_par_var_id):
            raise ValueError(
                f"Cumulative variable {cumulative_par_var_id} not present in parent's {component_par.id}"
            )

        elif not component_chd.is_variable_in_model(cumulative_chd_var_id):
            raise ValueError(
                f"Cumulative variable {cumulative_chd_var_id} not present in {component_chd.id}"
            )

        elif not component_chd.is_variable_in_model(incremental_chd_var_id):
            raise ValueError(
                f"Incremental variable {incremental_chd_var_id} not present in {component_chd.id}"
            )

        prefix_par = ""
        cumulative_var_par_id = ""
        if self.parent:
            prefix_par = f"{self.parent.id}.{component_par.id}"
            cumulative_var_par_id = f"{prefix_par}.{cumulative_par_var_id}"

            self.coupling_info.variables.add(cumulative_var_par_id)

        prefix_chd = f"{self.id}.{component_chd.id}"
        cumulative_var_chd_id = f"{prefix_chd}.{cumulative_chd_var_id}"
        incremental_var_par_id = f"{prefix_chd}.{incremental_chd_var_id}"

        self.coupling_info.variables.add(cumulative_var_chd_id)
        self.coupling_info.variables.add(incremental_var_par_id)

        self.coupling_info.constraints[
            f"{prefix_par}.{prefix_chd}.{len(self.coupling_info.constraints)}"
        ] = (var(cumulative_var_chd_id) - var(incremental_var_par_id)) == (
            var(cumulative_var_par_id) if cumulative_var_par_id else literal(0)
        )

    def _add_coupling_constraints(self) -> None:
        variables: List[Variable] = []
        constraints: List[Constraint] = []

        """
        Here and below, we use (and mostly abuse!) of the fact that
        both coupling_network and coupling_info attributes are initialized,
        or bounded, at the function definition since they are default arguments!
        It allows us to not iterate over the tree, since all nodes share the same
        coupling_network and coupling_info objects.

        However, IT IS BAD DESIGN!
        But for now it will suffice since the inner workings will change once
        Thomas has changed the AST instantiation
        TODO Update this once the new AST is merged
        """
        for var_name in self.coupling_info.variables:
            variables.append(
                # TODO For now, unbounded positive float variable
                float_variable(
                    var_name,
                    lower_bound=literal(0),
                    structure=IndexingStructure(False, False),
                    context=ProblemContext.INVESTMENT,
                ),
            )

        for cst_name, expr in self.coupling_info.constraints.items():
            constraints.append(
                Constraint(
                    name=cst_name,
                    expression=expr,
                    context=ProblemContext.INVESTMENT,
                ),
            )

        self.coupling_network.add_component(
            create_component(
                model(
                    id="",
                    variables=variables,
                    constraints=constraints,
                ),
                id="coupling",
            )
        )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List

from gems.expression.indexing_structure import IndexingStructure

from .expression import (
    AdditionNode,
    AllTimeSumNode,
    ComparisonNode,
    ComponentParameterNode,
    ComponentVariableNode,
    DivisionNode,
    ExpressionNode,
    LiteralNode,
    MultiplicationNode,
    NegationNode,
    ParameterNode,
    PortFieldAggregatorNode,
    PortFieldNode,
    ProblemParameterNode,
    ProblemVariableNode,
    ScenarioOperatorNode,
    TimeEvalNode,
    TimeShiftNode,
    TimeSumNode,
    VariableNode,
)
from .visitor import ExpressionVisitor, T, visit


class IndexingStructureProvider(ABC):
    @abstractmethod
    def get_parameter_structure(self, name: str) -> IndexingStructure:
        ...

    @abstractmethod
    def get_variable_structure(self, name: str) -> IndexingStructure:
        ...

    @abstractmethod
    def get_component_variable_structure(
        self, component_id: str, name: str
    ) -> IndexingStructure:
        ...

    @abstractmethod
    def get_component_parameter_structure(
        self, component_id: str, name: str
    ) -> IndexingStructure:
        ...


@dataclass(frozen=True)
class TimeScenarioIndexingVisitor(ExpressionVisitor[IndexingStructure]):
    """
    Determines if the expression represents a single expression or an expression that should be instantiated for all time steps.
    """

    context: IndexingStructureProvider

    def literal(self, node: LiteralNode) -> IndexingStructure:
        return IndexingStructure(False, False)

    def negation(self, node: NegationNode) -> IndexingStructure:
        return visit(node.operand, self)

    def _combine(self, operands: List[ExpressionNode]) -> IndexingStructure:
        if not operands:
            return IndexingStructure(False, False)
        res = visit(operands[0], self)
        if res.is_time_scenario_varying():
            return res
        for o in operands[1:]:
            res = res | visit(o, self)
            if res.is_time_scenario_varying():
                return res
        return res

    def addition(self, node: AdditionNode) -> IndexingStructure:
        # performance note:
        # here we don't need to visit all nodes, we can stop as soon as
        # index is true/true
        return self._combine(node.operands)

    def multiplication(self, node: MultiplicationNode) -> IndexingStructure:
        return self._combine([node.left, node.right])

    def division(self, node: DivisionNode) -> IndexingStructure:
        return self._combine([node.left, node.right])

    def comparison(self, node: ComparisonNode) -> IndexingStructure:
        return self._combine([node.left, node.right])

    def variable(self, node: VariableNode) -> IndexingStructure:
        time = self.context.get_variable_structure(node.name).time == True
        scenario = self.context.get_variable_structure(node.name).scenario == True
        return IndexingStructure(time, scenario)

    def parameter(self, node: ParameterNode) -> IndexingStructure:
        time = self.context.get_parameter_structure(node.name).time == True
        scenario = self.context.get_parameter_structure(node.name).scenario == True
        return IndexingStructure(time, scenario)

    def comp_variable(self, node: ComponentVariableNode) -> IndexingStructure:
        return self.context.get_component_variable_structure(
            node.component_id, node.name
        )

    def comp_parameter(self, node: ComponentParameterNode) -> IndexingStructure:
        return self.context.get_component_parameter_structure(
            node.component_id, node.name
        )

    def pb_variable(self, node: ProblemVariableNode) -> IndexingStructure:
        raise ValueError(
            "Not relevant to compute indexation on already instantiated problem variables."
        )

    def pb_parameter(self, node: ProblemParameterNode) -> IndexingStructure:
        raise ValueError(
            "Not relevant to compute indexation on already instantiated problem parameters."
        )

    def time_shift(self, node: TimeShiftNode) -> IndexingStructure:
        return visit(node.operand, self)

    def time_eval(self, node: TimeEvalNode) -> IndexingStructure:
        return IndexingStructure(False, visit(node.operand, self).scenario)

    def time_sum(self, node: TimeSumNode) -> IndexingStructure:
        return visit(node.operand, self)

    def all_time_sum(self, node: AllTimeSumNode) -> IndexingStructure:
        return IndexingStructure(False, visit(node.operand, self).scenario)

    def scenario_operator(self, node: ScenarioOperatorNode) -> IndexingStructure:
        return IndexingStructure(visit(node.operand, self).time, False)

    def port_field(self, node: PortFieldNode) -> IndexingStructure:
        raise ValueError(
            "Port fields must be resolved before computing indexing structure."
        )

    def port_field_aggregator(self, node: PortFieldAggregatorNode) -> IndexingStructure:
        raise ValueError(
            "Port fields aggregators must be resolved before computing indexing structure."
        )


def compute_indexation(
    expression: ExpressionNode, provider: IndexingStructureProvider
) -> IndexingStructure:
    return visit(expression, TimeScenarioIndexingVisitor(provider))
